# Design notes

MODOS is a WIP and might diverge from its initial design as requirements evolve.

## Initial design

The initial design is based on the following principles:

### Structure

A multi-omics digital object (MODO) contains:
* A zarr archive as a subfolder, responsible for storing metadata and potentially arrays.
* Any number of files at the root of the object.

### Metadata

Modos manages metadata using the Zarr API, which provides a pythonic abstraction over its hierarchically organized metadata and arrays. Metadata is stored as json files.

#### Strengths

* zarr is optimized for efficient access over s3, providing some useful features for modos.
* relying on the zarr api avoids tedious metadata manipulation codes in modos.

#### Weaknesses

* zarr arrays are not flexible enough to hold tabular data which is commonly encountered in omics (named axes, heterogenous and variable sized data types).


### Schema

The metadata follows a schema (see [modos-schema](https://github.com/sdsc-ordes/modos-schema)) which modos enforces when writing data to the zarr archive. MODOS relies on linkml to define and enforce the schema. In practice that means:

* modos-schema is defined in yaml using linkml's metaschema
* a companion python library (modos-schema) generated by linkml using metaprogramming (i.e. automatic code generation) provides parsing, (de)serialization and validation utilities to work with the metadata.

#### Strengths

* the schema has a single source of truth (in yaml) and all other serializations (rdf, jsonschema, protobuf) as well as library code is auto-generated. This drastically reduces schema maintenance work.
* introspection capabilities of generated code are quite powerful and allow modos-api to be more schema-independent.

#### Weaknesses

* speed is lackluster; the linkml-generated dataclasses take up most of modos runtime.
* linkml is prone to breakage and changes rapidly, while it has been very useful for prototyping, we may need to look for a more stable solution on the long term.


## Potential improvements

### Metadata

As we are not benefitting much from zarr arrays (we likely need to store tabular data as parquet files), and only use it for metadata management, there may be more appropriate frameworks for metadata management of digital objects. Some popular examples include:
* RO-crates: Single json-ld file describing the object contents using schema.org terminology
  + potential for simplifying modos, however we need to check how / if RO-crate terminology can be extended to support modos-schema
* biocompute objects: Rather specialized in workflow description, probably not the best fit.
* datapackages: focus on describing contents on files, but not contextual information (submitter, sample, ...)


Among the above frameworks, RO-crate seems the closest to modos' requirements. If it turns out not to be flexible enough, we could also replace it with a standard json(-ld) file following modos-schema. In such case, the structure could resemble the example below.:


```
.
├── modos.json  # ← metadata about object itself, update date, spec version, ...
├── consolidated_metadata.json  # ← auto-generated union of all object metadata (to minimize IO overhead over s3)
├── assay
│   └── assay1.json
├── data
│   ├── demo1.parquet
│   ├── demo1.json
│   ├── demo2.cram
│   └── demo2.json
├── reference
    │   ├── ref1.fasta
│   └── ref1.json
├── sample
│   └── sample1.json
└── sequence
    └── seq1.json
```

### Schema

There does not appear to be an alternative to linkml, therefore replacing it would introduce additional maintenance effort, to keep code definitions in line with the schema. In principle, the (de)serialization and validation code could be written as part of modos (making it more tightly coupled to modos-schema) using pydantic or a similar high performance library.

A hybrid solution (being explored) is to keep using linkml with their experimental pydantic code generator. While it should provide better performance, the generated pydantic code appears much less powerful than the generated dataclasses.
